package Bio::MUST::Drivers::Hmmer;
# ABSTRACT: Bio::MUST driver for running HMMER3 (mainly hmmbuild and hmmsearch)
# CONTRIBUTOR: Arnaud DI FRANCO <arnaud.difranco@gmail.com>
$Bio::MUST::Drivers::Hmmer::VERSION = '0.161550';
use Moose;
use namespace::autoclean;

use autodie;
use feature qw(say);

use Modern::Perl;
use Carp;
use IPC::System::Simple qw(system);
use File::Temp; my $template = 'tmpfile_XXXX';
use Path::Class;

use Bio::FastParsers::Hmmer::Standard;
use Bio::FastParsers::Hmmer::Table;
use Bio::FastParsers::Hmmer::DomTable;

use Bio::MUST::Core::Types;
use aliased 'Bio::MUST::Drivers::Hmmer::Model';
use aliased 'Bio::MUST::Core::Ali';
use Bio::MUST::Drivers::Utils qw(stringify_args);

use Smart::Comments -ENV;


has 'ali' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::Ali',
    required => 1,
    coerce   => 1,
);

# TODO: move to FastParsers
has 'model' => (
    is      => 'ro',
    isa     => 'Maybe[Bio::MUST::Drivers::Hmmer::Model]',
    lazy    => 1,                   # or Bio::MUST::Core::Types::Files?
    builder => '_build_model',
);

has 'consider_X' => (
    is      => 'ro',
    isa     => 'Bool',
    default => 1,
);

# TODO: check how to avoid this
has 'debug_mode' => (
    is      => 'ro',
    isa     => 'Bool',
    default => 0,
);

# TODO: think about API here: do we really want this? maybe yes
has 'args' => (
    traits  => ['Hash'],
    is      => 'ro',
    isa     => 'HashRef',
    default => sub{ {} },
);

## no critic (ProhibitUnusedPrivateSubroutines)

sub _build_model {
    my $self = shift;

    my $ali = $self->ali;
    ##### N seqs in ali: $ali->count_seqs

    # skip model creation if empty Ali
    return unless $ali->count_seqs;

    # setup input/output files
    my $tmpfasta = File::Temp->new( TEMPLATE => $template,
        SUFFIX => '_model.fasta', UNLINK => $self->debug_mode ? 0 : 1 );
    # TODO: check how to avoid this
    $ali->gapify_seqs( $self->consider_X ? 'X' : undef );
    $ali->store_fasta($tmpfasta);
    my $hmm_model = File::Temp-> new( TEMPLATE => $template, SUFFIX => '.hmm',
        UNLINK => $self->debug_mode ? 0 : 1 );
    my $hmm_out   = File::Temp-> new( TEMPLATE => $template, SUFFIX => '.out');

    my $args = $self->args;

    # TODO: check the comment below
    # seems to be not so reliable -> so it is returning 0 and failed dna
    $args->{ $ali->is_protein ? '--amino' : '--dna' } = undef;
    ###### $args
    my $args_str = stringify_args($args);
    ###### $args_str

    # WHAT THE HELL with IPC::SYSTEM
    # I have to set $ENV{PATH} or I'll get a error
    # the bug does not seem to occur anymore
    #~ $ENV{PATH} = '/usr/local/bin/';

    # create hmmbuild command
    my $pgm = 'hmmbuild';
    my $cmd = "$pgm $args_str $hmm_model $tmpfasta >> $hmm_out";
    #### $cmd

    # try to robustly execute hmmbuild
    my $ret_code = system( [ 0, 127 ], $cmd);
    if ($ret_code == 127) {
        carp "Cannot execute $pgm command; returning without HMM model!";
        return;
    }

    return Bio::MUST::Drivers::Hmmer::Model->new( file => $hmm_model );
}

## use critic


sub search {
    my $self    = shift;
    my $targets = shift;
    my $args    = shift // {};

    # skip search if empty Ali (undefined model)
    return unless defined $self->model;

    # setup input/output files
    my $hmm_model = $self->model->file;
    my $target = File::Temp->new($template, SUFFIX => '_target.fasta',
        UNLINK => $self->debug_mode ? 0 : 1 );
    my $temp_ali = Ali->new( seqs => $targets );
    $temp_ali->store_fasta( $target, { degap => 1 } );
    my $res    = File::Temp->new($template, SUFFIX => '.res',
        UNLINK => $self->debug_mode ? 0 : 1 );
    my $out    = File::Temp->new($template, SUFFIX => '.out');

    # setup output file format and parser subclass
    my $parser_class;
    if (exists $args->{'--notextw'}) {
        $args->{'-o'} = $res;
        $parser_class = 'Bio::FastParsers::Hmmer::Standard';
    } elsif (exists $args->{'--domtblout'}) {
        $args->{'--domtblout'} = $res;
        $parser_class = 'Bio::FastParsers::Hmmer::DomTable';
    } elsif (exists $args->{'--tblout'}) {
        $args->{'--tblout'} = $res;
        $parser_class = 'Bio::FastParsers::Hmmer::Table';
    }
    return unless defined $parser_class;

    # create hmmsearch command
    my $args_str = stringify_args($args);
    my $pgm = 'hmmsearch';
    my $cmd = "$pgm $args_str $hmm_model $target >> $out 2>> $out";
    # TODO: clarify the use of $out and of the appending mode

    # try to robustly execute hmmsearch
    my $ret_code = system( [ 0, 127 ], $cmd);
    if ($ret_code == 127) {
        carp 'Warning: cannot execute $pgm command; returning without parser!';
        return;
    }

    my $parser = $parser_class->new( file => file($res) );
    return $parser;
}

__PACKAGE__->meta->make_immutable;
1;

__END__

=pod

=head1 NAME

Bio::MUST::Drivers::Hmmer - Bio::MUST driver for running HMMER3 (mainly hmmbuild and hmmsearch)

=head1 VERSION

version 0.161550

=head1 SYNOPSIS

    # TODO

=head1 DESCRIPTION

    # TODO

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
